#include "matchcompiler.h"
#include <string>
#include <cstring>
#include "errorlogger.h"
#include "token.h"
// pattern: %var% [ %num%| ] = {
static inline bool match1(const Token* tok) {
    if (!tok || !(tok->varId() != 0))
        return false;
    tok = tok->next();
    if (!tok || !((tok->tokType() == Token::eExtendedOp || tok->tokType() == Token::eLambda) && tok->str() == MatchCompiler::makeConstString("[")))
        return false;
    tok = tok->next();
    if (tok && (tok->isNumber()))
        tok = tok->next();
    if (!tok || !((tok->tokType() == Token::eExtendedOp || tok->tokType() == Token::eLambda) && tok->str() == MatchCompiler::makeConstString("]")))
        return false;
    tok = tok->next();
    if (!tok || !((tok->tokType() == Token::eAssignmentOp) && tok->str() == MatchCompiler::makeConstString("=")))
        return false;
    tok = tok->next();
    if (!tok || !((tok->tokType() == Token::eBracket) && tok->str() == MatchCompiler::makeConstString("{")))
        return false;
    return true;
}
// pattern: =
static inline bool match2(const Token* tok) {
    if (!tok || !((tok->tokType() == Token::eAssignmentOp) && tok->str() == MatchCompiler::makeConstString("=")))
        return false;
    return true;
}
// pattern: const %type% %var% [ %num%| ] = {
static inline bool match3(const Token* tok) {
    if (!tok || !((tok->tokType() == Token::eKeyword) && tok->str() == MatchCompiler::makeConstString("const")))
        return false;
    tok = tok->next();
    if (!tok || !(tok->isName() && tok->varId() == 0U))
        return false;
    tok = tok->next();
    if (!tok || !(tok->varId() != 0))
        return false;
    tok = tok->next();
    if (!tok || !((tok->tokType() == Token::eExtendedOp || tok->tokType() == Token::eLambda) && tok->str() == MatchCompiler::makeConstString("[")))
        return false;
    tok = tok->next();
    if (tok && (tok->isNumber()))
        tok = tok->next();
    if (!tok || !((tok->tokType() == Token::eExtendedOp || tok->tokType() == Token::eLambda) && tok->str() == MatchCompiler::makeConstString("]")))
        return false;
    tok = tok->next();
    if (!tok || !((tok->tokType() == Token::eAssignmentOp) && tok->str() == MatchCompiler::makeConstString("=")))
        return false;
    tok = tok->next();
    if (!tok || !((tok->tokType() == Token::eBracket) && tok->str() == MatchCompiler::makeConstString("{")))
        return false;
    return true;
}
// pattern: const char %var% [ %num%| ] = %str% ;
static inline bool match4(const Token* tok) {
    if (!tok || !((tok->tokType() == Token::eKeyword) && tok->str() == MatchCompiler::makeConstString("const")))
        return false;
    tok = tok->next();
    if (!tok || !(tok->str() == MatchCompiler::makeConstString("char")))
        return false;
    tok = tok->next();
    if (!tok || !(tok->varId() != 0))
        return false;
    tok = tok->next();
    if (!tok || !((tok->tokType() == Token::eExtendedOp || tok->tokType() == Token::eLambda) && tok->str() == MatchCompiler::makeConstString("[")))
        return false;
    tok = tok->next();
    if (tok && (tok->isNumber()))
        tok = tok->next();
    if (!tok || !((tok->tokType() == Token::eExtendedOp || tok->tokType() == Token::eLambda) && tok->str() == MatchCompiler::makeConstString("]")))
        return false;
    tok = tok->next();
    if (!tok || !((tok->tokType() == Token::eAssignmentOp) && tok->str() == MatchCompiler::makeConstString("=")))
        return false;
    tok = tok->next();
    if (!tok || !(tok->tokType() == Token::eString))
        return false;
    tok = tok->next();
    if (!tok || !(tok->str() == MatchCompiler::makeConstString(";")))
        return false;
    return true;
}
#line 1 "/builddir/build/BUILD/cppcheck-2.16.2-build/cppcheck-2.16.2/lib/vf_array.cpp"
/*
 * Cppcheck - A tool for static C/C++ code analysis
 * Copyright (C) 2007-2024 Cppcheck team.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "vf_array.h"

#include "astutils.h"
#include "config.h"
#include "symboldatabase.h"
#include "token.h"
#include "tokenlist.h"
#include "vfvalue.h"

#include "vf_settokenvalue.h"

#include <map>
#include <utility>

namespace ValueFlow
{
    void analyzeArray(TokenList &tokenlist, const Settings &settings)
    {
        std::map<nonneg int, const Token *> constantArrays;

        for (Token *tok = tokenlist.front(); tok; tok = tok->next()) {
            if (tok->varId() > 0) {
                // array
                const std::map<nonneg int, const Token *>::const_iterator it = constantArrays.find(tok->varId());
                if (it != constantArrays.end()) {
                    Value value;
                    value.valueType = Value::ValueType::TOK;
                    value.tokvalue = it->second;
                    value.setKnown();
                    setTokenValue(tok, std::move(value), settings);
                }

                // const array decl
                else if (tok->variable() && tok->variable()->isArray() && tok->variable()->isConst() &&
                         tok->variable()->nameToken() == tok && match1(tok)) {
                    Token* rhstok = tok->linkAt(1)->tokAt(2);
                    constantArrays[tok->varId()] = rhstok;
                    tok = rhstok->link();
                }

                // pointer = array
                else if (tok->variable() && tok->variable()->isArray() && match2(tok->astParent()) &&
                         astIsRHS(tok) && tok->astParent()->astOperand1() &&
                         tok->astParent()->astOperand1()->variable() &&
                         tok->astParent()->astOperand1()->variable()->isPointer()) {
                    Value value;
                    value.valueType = Value::ValueType::TOK;
                    value.tokvalue = tok;
                    value.setKnown();
                    setTokenValue(tok, std::move(value), settings);
                }
                continue;
            }

            if (match3(tok)) {
                Token *vartok = tok->tokAt(2);
                Token *rhstok = vartok->linkAt(1)->tokAt(2);
                constantArrays[vartok->varId()] = rhstok;
                tok = rhstok->link();
                continue;
            }

            if (match4(tok)) {
                Token *vartok = tok->tokAt(2);
                Token *strtok = vartok->linkAt(1)->tokAt(2);
                constantArrays[vartok->varId()] = strtok;
                tok = strtok->next();
                continue;
            }
        }
    }
}
